<?php
declare(strict_types=1);

namespace Inti\Wp\Provider\Plugin;

use Inti\Provider\Query\QueryInterface;
use Inti\Provider\Query\QueryPaginateInterface;
use Inti\Provider\Query\QueryParentInterface;
use Inti\Provider\Query\QuerySlugInterface;
use Inti\Provider\Service\ProviderManagerInterface;
use Inti\Provider\Type\TypeInterface;
use Inti\Wp\Plugin\PluginInterface;
use Inti\Wp\Provider\Service\IdManager;
use Inti\Wp\Provider\Service\SqlHelpers;
use Inti\Wp\Provider\Type\Term\TaxonomyInterface;
use WP_Term;
use WP_Term_Query;

use function add_filter;
use function array_map;
use function count;
use function is_array;
use function is_string;
use function register_taxonomy;

/**
 * Class ProvideTerms
 *
 * @package Inti\Wp\Provider\Plugin
 * @author Danijel Fabijan <hipnaba@gmail.com>
 * @link https://bitbucket.org/oktan-nautika/inti-wp-provider
 */
final class ProvideTerms implements PluginInterface
{
    use SqlHelpers;

    private ProviderManagerInterface $providers;
    private IdManager $ids;
    /** @var array<string, TypeInterface> */
    private array $provided = [];

    public function __construct(ProviderManagerInterface $providers, IdManager $idOffsets)
    {
        $this->providers = $providers;
        $this->ids = $idOffsets;
    }

    public function run(?array $options = null): void
    {
        $providers = $this->providers->getProviders();

        if (count($providers) === 0) {
            return;
        }

        add_action('init', [$this, 'register_taxonomies']);
        add_filter('query', [$this, 'query']);
        add_filter('terms_pre_query', [$this, 'terms_pre_query'], 10, 2);

        foreach ($providers as $provider) {
            $type = $provider->getProvidedType();
            $this->provided[$type->getName()] = $type;
        }
    }

    public function terms_pre_query(?array $terms, WP_Term_Query $query)
    {
        $q = &$query->query_vars;
        $taxonomy = is_array($q['taxonomy']) ? $q['taxonomy'][0] : $q['taxonomy'];

        if (!is_string($taxonomy) || empty($taxonomy) || !isset($this->provided[$taxonomy])) {
            return $terms;
        }

        $provider = $this->providers->getProvider($this->provided[$taxonomy]);

        if ($provider instanceof QueryInterface) {
            $provider = clone $provider;
            $provider->query();

            if ($provider instanceof QueryPaginateInterface && !empty($q['number'])) {
                $provider->paginate($q['offset'] / $q['number'] + 1, $q['number']);
            }

            if ($provider instanceof QueryParentInterface && !empty($q['parent'])) {
                $parent = $this->ids->isOffset($q['parent'])
                    ? $this->ids->unoffset($q['parent'])
                    : $q['parent'];
                $provider->parent($parent);
            }

            if ($provider instanceof QuerySlugInterface && !empty($q['slug'])) {
                $provider->slugs((array) $q['slug']);
            }

            if (!empty($q['include'])) {
                $ids = array_map('intval', $q['include']);
                $provider->ids(
                    array_map([$this->ids, 'unoffset'], $ids)
                );
            }

            $fields = $q['fields'] ?: 'all';

            if ($fields === 'count') {
                return $provider->getCount();
            }

            foreach ($provider->getItems() as $row) {
                $row = $this->prepareRow($row, $provider->getProvidedType());

                switch ($fields) {
                    case 'id=>parent':
                        $parent = $row['parent'] > 0
                            ? $this->ids->offset($row['parent'], $provider->getProvidedType())
                            : $row['parent'];

                        $terms[$row['term_id']] = $parent;
                        break;
                    case 'all':
                    case 'all_with_object_id':
                    default:
                        $terms[] = new WP_Term((object) $row);
                }
            }
        }

        return $terms;
    }

    public function query(string $query): string
    {
        if (preg_match('/^SELECT t\.\*, tt\.\* FROM wp_terms AS t INNER JOIN wp_term_taxonomy AS tt ON t\.term_id = tt\.term_id WHERE t\.term_id = (\d+)$/', $query, $matches)) {
            $term_id = intval($matches[1]);

            if ($this->ids->isOffset($term_id)) {
                /** @var TaxonomyInterface $taxonomy */
                $taxonomy = $this->ids->getType($term_id);
                $term_id = $this->ids->unoffset($term_id);

                if (($provider = $this->providers->getProvider($taxonomy)) && $provider instanceof QueryInterface) {
                    $provider->query()
                        ->id($term_id);

                    if ($row = $provider->getItem()) {
                        $row = $this->prepareRow($row, $taxonomy);
                        $query = $this->arrayToSelect($row);
                    }
                }
            }
        }

        return $query;
    }

    private function prepareRow(array $row, TaxonomyInterface $type): array
    {
        if (isset($row['parent']) && $row['parent'] > 0) {
            $parentTaxonomy = $type->getParentTaxonomy();
            $parent = $this->ids->offset($row['parent'], $parentTaxonomy);
        }

        $row['term_id'] = $this->ids->offset(intval($row['term_id']), $type);
        $row['name'] = $row['name'] ?? '';
        $row['slug'] = $row['slug'] ?? '';
        $row['term_group'] = 0;
        $row['term_taxonomy_id'] = $row['term_id'];
        $row['taxonomy'] = $type->getName();
        $row['description'] = $row['description'] ?? '';
        $row['parent'] = $parent ?? 0;
        $row['count'] = $row['count'] ?? 0;

        return $row;
    }

    public function register_taxonomies(): void
    {
        foreach ($this->providers->getProviders() as $provider) {
            $type = $provider->getProvidedType();
            add_filter("pre_option_{$type->getName()}_children", '__return_false');

            if ($type instanceof TaxonomyInterface && !$type->isBuiltin()) {
                $this->ids->registerOffset($type);

                register_taxonomy($type->getName(), $type->getAssociatedType()->getName(), [
                    'description' => $type->getDescription(),
                    'hierarchical' => $type->isHierarchical(),
                    'labels' => $type->getLabels(),
                    'public' => $type->isPublic(),
                    'publicly_queryable' => $type->isPubliclyQueryable(),
                    'show_ui' => $type->showUi(),
                    'show_in_menu' => $type->showInMenu(),
                    'show_in_nav_menus' => $type->showInNavMenus(),
                    'show_tagcloud' => $type->showTagCloud(),
                    'show_in_quick_edit' => $type->showInQuickEdit(),
                    'show_admin_column' => $type->showAdminColumn(),
                    'meta_box_cb' => $type->getMetaBoxCallback(),
                    'meta_box_sanitize_cb' => null,
                    'capabilities' => $type->getCapabilities(),
                    'rewrite' => $type->getRewriteConfig(),
                    'query_var' => $type->getQueryVar(),
                    'update_count_callback' => '',
                    'show_in_rest' => false,
                    'rest_base' => false,
                    'rest_controller_class' => false,
                    'default_term' => null,
                ]);
            }
        }
    }
}
