<?php
declare(strict_types=1);

namespace Inti\Wp\Provider\Service;

use Inti\Provider\Type\TypeInterface;

use function function_exists;
use function get_class;
use function get_option;
use function intval;
use function is_callable;
use function max;
use function update_option;

/**
 * Class IdManager
 *
 * @package Inti\Wp\Provider\Service
 * @author Danijel Fabijan <hipnaba@gmail.com>
 * @link https://bitbucket.org/oktan-nautika/inti-wp-provider
 */
final class IdManager
{
    const BASE_OFFSET = 1e8;
    const OPTION_NAME = 'inti_ids';

    private array $offsets = [];

    public static function instance(): IdManager
    {
        static $instance;

        if (!isset($instance)) {
            $instance = new self();
        }

        return $instance;
    }

    private function __construct()
    {
        if (function_exists('get_option')) {
            $this->offsets = get_option(self::OPTION_NAME, []);
        }
    }

    public function isOffset(int $id): bool
    {
        return $this->getType($id) instanceof TypeInterface;
    }

    public function offset(int $id, TypeInterface $type): int
    {
        return intval($id + ($this->hasOffset($type) ? $this->getOffset($type) : $this->registerOffset($type)));
    }

    public function unoffset(int $id): int
    {
        return $this->isOffset($id)
            ? intval($id - $this->getOffset($this->getType($id)))
            : $id;
    }

    public function hasOffset(TypeInterface $type): bool
    {
        return isset($this->offsets[get_class($type)]);
    }

    public function getOffset(TypeInterface $type): int
    {
        return $this->offsets[get_class($type)] ?? 0;
    }

    public function registerOffset(TypeInterface $type): int
    {
        if ($this->hasOffset($type)) {
            return $this->getOffset($type);
        }

        $highest = count($this->offsets) > 0 ? max($this->offsets) : 0;
        $newOffset = intval($highest + self::BASE_OFFSET);

        $this->offsets[get_class($type)] = $newOffset;

        if (function_exists('update_option')) {
            update_option(self::OPTION_NAME, $this->offsets);
        }

        return $newOffset;
    }

    public function getType(int $id): ?TypeInterface
    {
        if ($id < self::BASE_OFFSET) {
            return null;
        }

        $offset = intval(floor($id / self::BASE_OFFSET) * self::BASE_OFFSET);
        $offsets = array_flip($this->offsets ?? []);

        if (!isset($offsets[$offset])) {
            return null;
        }

        $factory = [$offsets[$offset], 'instance'];
        return is_callable($factory) ? $factory() : null;
    }


}
